import { BehaviorSubject, of, Subject, from } from 'rxjs';
import { concatMap, delay, merge } from 'rxjs/operators';
import { APP_BASE_HREF, Location, LocationStrategy, PlatformLocation, CommonModule } from '@angular/common';
import { Injectable, NgModuleFactoryLoader, InjectionToken, Compiler, ComponentFactoryResolver, Injector, Inject, Optional, Attribute, Component, EventEmitter, Input, Output, ViewChild, ViewContainerRef, Directive, NgModule, APP_BOOTSTRAP_LISTENER, ApplicationRef, ANALYZE_FOR_ENTRY_COMPONENTS, SystemJsNgModuleLoader } from '@angular/core';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * @abstract
 */
class RouterModuleFactory {
}
class DefaultRouterModuleFactory extends RouterModuleFactory {
    /**
     * @param {?} loader
     */
    constructor(loader) {
        super();
        this.loader = loader;
        this.loaded = {};
    }
    /**
     * @param {?} moduleName
     * @param {?} parentInjector
     * @return {?}
     */
    getModule(moduleName, parentInjector) {
        if (this.loaded[moduleName]) {
            return Promise.resolve(this.loaded[moduleName]);
        }
        return this.loader.load(moduleName).then(factory => {
            return this.loaded[moduleName] = factory.create(parentInjector);
        });
    }
}
DefaultRouterModuleFactory.decorators = [
    { type: Injectable }
];
/** @nocollapse */
DefaultRouterModuleFactory.ctorParameters = () => [
    { type: NgModuleFactoryLoader }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class UrlParser {
    /**
     * @return {?}
     */
    static ensureParser() {
        if (!UrlParser.parser) {
            UrlParser.parser = document.createElement("a");
        }
    }
    /**
     * @param {?} url
     * @return {?}
     */
    static parseUrl(url) {
        UrlParser.ensureParser();
        UrlParser.parser.href = url;
        // Host, Hostname, Port Protocol могут быть пучтыми в IE на относительных URL
        /** @type {?} */
        const parseResult = {
            hash: UrlParser.parser.hash,
            host: UrlParser.parser.host,
            hostname: UrlParser.parser.hostname,
            pathname: UrlParser.parser.pathname[0] == "/" ? UrlParser.parser.pathname : "/" + UrlParser.parser.pathname,
            //IE FIX
            port: UrlParser.parser.port,
            protocol: UrlParser.parser.protocol,
            search: UrlParser.parser.search
        };
        UrlParser.parser.href = "";
        return parseResult;
    }
}
UrlParser.parser = null;

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * @param {?} url
 * @param {?} route
 * @return {?}
 */
function defaultRouteMatcher(url, route) {
    //drop query string for matching
    url = UrlParser.parseUrl(url).pathname;
    /** @type {?} */
    let parsedRoute = parseRoute(route.path);
    /** @type {?} */
    let parsedUrl = url.match(parsedRoute.regexp);
    if (parsedUrl) {
        /** @type {?} */
        let params = {};
        for (let i = 0; i < parsedRoute.paramsMap.length; i++) {
            params[parsedRoute.paramsMap[i].name] = parsedUrl[1 + i];
        }
        return Promise.resolve({
            component: route.component,
            params,
            route
        });
    }
    return Promise.resolve(null);
}
/**
 * Normalizes the given route template string, returning a regular expression representing a route
 * and the original path string.
 *
 * Inspired by pathRexp in visionmedia/express/lib/utils.js.
 * and angular/angular.js/blob/master/src/ngRoute/route.js
 * @param {?} routeUrl
 * @return {?}
 */
function parseRoute(routeUrl) {
    /** @type {?} */
    let params = [];
    /** @type {?} */
    let result = {
        path: routeUrl,
        regexp: null,
        paramsMap: params
    };
    /* since our route table does not apply prefixes to url via forChild
     *  we should assume all routes start from root */
    if (routeUrl[0] !== "/") {
        routeUrl = "/" + routeUrl;
    }
    /** @type {?} */
    let path = routeUrl
        .replace(/([().])/g, "\\$1") // escape round brackets and dot with backslash
        .replace(/(\/)?:(\w+)(\*\?|[?*])?/g, function (_, slash, key, option) {
        // find parameters and    wildcards in string
        /** @type {?} */
        let optional = (option === "?" || option === "*?") ? "?" : null;
        /** @type {?} */
        let star = (option === "*" || option === "*?") ? "*" : null;
        params.push({ name: key, optional: !!optional });
        slash = slash || "";
        return ""
            + (optional ? "" : slash)
            + "(?:"
            + (optional ? slash : "")
            + (star && "(.+?)" || "([^/]+)")
            + (optional || "")
            + ")"
            + (optional || "");
    })
        .replace(/([/$*])/g, "\\$1");
    result.regexp = new RegExp("^" + path + "\/?$"); // support trailing slash
    return result;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const ROUTE_CONFIG = new InjectionToken("router ex route config");
/** @enum {number} */
const ReuseRouteStrategy = {
    /**
     * Create only one instance of route component and allways reuse them
     */
    STICKY: 1,
    /**
     * Reuse route component only when use return on the route by history pop event
     */
    CACHEBACK: 2,
};
ReuseRouteStrategy[ReuseRouteStrategy.STICKY] = 'STICKY';
ReuseRouteStrategy[ReuseRouteStrategy.CACHEBACK] = 'CACHEBACK';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * @param {?} loader
 * @param {?} parentInjector
 * @param {?} compiler
 * @param {?} resolve
 * @return {?}
 */
function createAsyncMatcher(loader, parentInjector, compiler, resolve) {
    return (url, route) => {
        if (!route.loadModule) {
            throw new Error("not an async route");
        }
        if (route.loadModuleCondition && !route.loadModuleCondition(url)) {
            return Promise.resolve(null); // don't need to load child module for specified url
        }
        if (typeof route.loadModule === "string") {
            return loader.getModule((/** @type {?} */ (route.loadModule)), parentInjector).then(ref => {
                return resolve(ref.injector.get(ROUTE_CONFIG), ref.injector, ref.componentFactoryResolver);
            });
        }
        else {
            /** @type {?} */
            const offlineMode = compiler instanceof Compiler;
            /** @type {?} */
            const loadModule = (/** @type {?} */ (route.loadModule));
            return loadModule()
                .then(module => {
                return offlineMode ? module : compiler.compileModuleAsync((/** @type {?} */ (module)));
            })
                .then((factory) => {
                /** @type {?} */
                const ref = factory.create(parentInjector);
                return resolve(ref.injector.get(ROUTE_CONFIG), ref.injector, ref.componentFactoryResolver);
            });
        }
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const ERROR_PAGE_PATH = "##";
/** @type {?} */
const NOTFOUND_PAGE_PATH = "**";
class RouteMatchService {
    /**
     * @param {?} resolver
     * @param {?} injector
     * @param {?} moduleLoader
     * @param {?} compiler
     */
    constructor(resolver, injector, moduleLoader, compiler) {
        this.resolver = resolver;
        this.injector = injector;
        this.moduleLoader = moduleLoader;
        this.compiler = compiler;
    }
    /**
     * @param {?} url
     * @param {?} routes
     * @return {?}
     */
    findRoute(url, routes) {
        return this.resolveInternal(url, routes, this.injector, this.resolver);
    }
    /**
     * @param {?} url
     * @param {?} routes
     * @param {?} injector
     * @param {?} resolver
     * @return {?}
     */
    resolveInternal(url, routes, injector, resolver) {
        return new Promise((resolve) => {
            if (url[0] !== "/") {
                url = "/" + url;
            }
            /** @type {?} */
            let index = 0;
            /** @type {?} */
            let processNext = () => {
                /** @type {?} */
                const matcher = this.createMatcher(url, routes[index], injector);
                matcher(url, routes[index])
                    .then((result) => {
                    if (!result) {
                        index++;
                        if (index >= routes.length) {
                            /** @type {?} */
                            const defaultRoute = routes.find(r => r.path === NOTFOUND_PAGE_PATH);
                            if (defaultRoute) {
                                /** @type {?} */
                                const notFoundResult = {
                                    route: defaultRoute,
                                    component: defaultRoute.component,
                                    injector: this.injector,
                                    factoryResolver: this.resolver,
                                    params: {}
                                };
                                resolve(notFoundResult); // no route found
                            }
                            else {
                                resolve(null);
                            }
                        }
                        else {
                            processNext();
                        }
                    }
                    else {
                        // from async modules matcher can return already resolved routes,
                        // if it is so, use original route from nested resolved route,
                        // not the original form top-level.
                        /** @type {?} */
                        const route = result.route || routes[index];
                        resolve({
                            params: result.params,
                            component: result.component,
                            factoryResolver: result.factoryResolver || resolver,
                            injector: result.injector || injector,
                            route: route
                        });
                    }
                }, err => {
                    /** @type {?} */
                    const errorRoute = routes.find(r => r.path === ERROR_PAGE_PATH);
                    if (errorRoute) {
                        /** @type {?} */
                        const errorResult = {
                            route: errorRoute,
                            component: errorRoute.component,
                            injector: this.injector,
                            factoryResolver: this.resolver,
                            params: {},
                            noCache: true
                        };
                        resolve(errorResult); // no route found
                    }
                    else {
                        resolve(null);
                    }
                    console.error(err);
                });
            };
            processNext();
        });
    }
    /**
     * @param {?} url
     * @param {?} route
     * @param {?} injector
     * @return {?}
     */
    createMatcher(url, route, injector) {
        if (route.matcher) {
            return route.matcher;
        }
        return route.loadModule ? createAsyncMatcher(this.moduleLoader, injector, this.compiler, (loadedRoutes, injector, resolver) => this.resolveInternal(url, loadedRoutes, injector, resolver)) : defaultRouteMatcher;
    }
}
RouteMatchService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
RouteMatchService.ctorParameters = () => [
    { type: ComponentFactoryResolver },
    { type: Injector },
    { type: RouterModuleFactory },
    { type: Compiler }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const PRIMARY_OUTLET = "__router-ex-default";
class RouterOutletMap {
    constructor() {
        this.map = {};
    }
    /**
     * @param {?} outlet
     * @param {?=} name
     * @return {?}
     */
    register(outlet, name) {
        this.map[name || PRIMARY_OUTLET] = outlet;
    }
    /**
     * @param {?} name
     * @return {?}
     */
    unregister(name) {
        delete this.map[name];
    }
    /**
     * @param {?=} name
     * @return {?}
     */
    getOutlet(name) {
        return this.map[name || PRIMARY_OUTLET];
    }
}
RouterOutletMap.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class NavigationStart {
    constructor() {
    }
}
class NavigationEnd {
    /**
     * @param {?} url
     * @param {?} component
     * @param {?} routeCtx
     */
    constructor(url, component, routeCtx) {
        this._url = url;
        this._component = component;
        this._routeCtx = routeCtx;
    }
    /**
     * @return {?}
     */
    get url() {
        return this._url;
    }
    /**
     * @return {?}
     */
    get activatedComponent() {
        return this._component;
    }
    /**
     * @return {?}
     */
    get routeContext() {
        return this._routeCtx;
    }
}
class NavigationCancel {
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class QueryStringParser {
    /**
     * @param {?} p
     * @return {?}
     */
    static serialize(p) {
        /** @type {?} */
        let res = [];
        for (let key in p) {
            if (p.hasOwnProperty(key)) {
                if (Array.isArray(p[key])) {
                    ((/** @type {?} */ (p[key]))).forEach((val) => {
                        res.push(key + (val === true ? "" : "=" + encodeURIComponent((/** @type {?} */ (val)))));
                    });
                }
                else {
                    if (!isNullOrUndefined(p[key])) {
                        res.push(key + (p[key] === true ? "" : "=" + encodeURIComponent((/** @type {?} */ (p[key])))));
                    }
                }
            }
        }
        return res.join("&");
    }
    /**
     * @param {?} rawQuery
     * @return {?}
     */
    static parse(rawQuery) {
        /** @type {?} */
        let params = {};
        if (!rawQuery) {
            return params;
        }
        if (rawQuery[0] === '?') {
            rawQuery = rawQuery.substring(1);
        }
        rawQuery.split("&").forEach((pair) => {
            let [key, value] = pair.split("=");
            if (value) {
                try {
                    value = decodeURIComponent(value);
                }
                catch (e) {
                    // drop invalid parameters
                    return;
                }
            }
            if (!params[key]) { // no such key yet
                params[key] = value || true;
            }
            else {
                if (Array.isArray(params[key])) { // value already an array
                    ((/** @type {?} */ (params[key]))).push(value || true);
                }
                else { // make an array
                    params[key] = ((/** @type {?} */ ([params[key], value || true])));
                }
            }
        });
        return params;
    }
}
/**
 * @param {?} obj
 * @return {?}
 */
function isNullOrUndefined(obj) {
    return typeof (obj) === "undefined" || obj === null;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * @param {?} pars1
 * @param {?} pars2
 * @param {?=} isRouteParams
 * @return {?}
 */
function isParamsEqual(pars1, pars2, isRouteParams) {
    /** @type {?} */
    let keys1 = Object.keys(pars1);
    /** @type {?} */
    let keys2 = Object.keys(pars2);
    if (!isRouteParams) {
        // remove parameters with null values
        keys1.forEach((k, i) => {
            if (pars1[k] === null) {
                delete pars1[k];
                keys1.splice(i, 1);
            }
        });
        keys2.forEach((k, i) => {
            if (pars2[k] === null) {
                delete pars2[k];
                keys2.splice(i, 1);
            }
        });
    }
    // quantity of parameters mismatch
    if (keys1.length !== keys2.length) {
        return false;
    }
    // different param names
    if (!keys1.every(k => keys2.indexOf(k) > -1)) {
        return false;
    }
    for (let key of keys1) {
        if (Array.isArray(pars1[key])) {
            // param is array in one set, not array in another
            if (!Array.isArray(pars2[key])) {
                return false;
            }
            // check nested arrays lengths and values
            if (((/** @type {?} */ (pars1[key]))).length
                == ((/** @type {?} */ (pars2[key]))).length
                && !((/** @type {?} */ (pars1[key]))).every((v) => ((/** @type {?} */ (pars2[key]))).indexOf(v) > -1)) {
                return false;
            }
        }
        else {
            // check simple non-array values
            if (pars1[key] !== pars2[key]) {
                return false;
            }
        }
    }
    // all checks passed, params sets are equal
    return true;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class Router {
    /**
     * @param {?} outletMap
     * @param {?} matchService
     * @param {?} location
     * @param {?} injector
     * @param {?} baseHref
     */
    constructor(outletMap, matchService, location, injector, baseHref) {
        this.outletMap = outletMap;
        this.matchService = matchService;
        this.location = location;
        this.injector = injector;
        this.baseHref = baseHref;
        this.resolvedRoutes = {};
        this.navigationId = 0;
        this.navigations = new BehaviorSubject(null);
        this._events = new Subject();
        if (!this.baseHref) {
            this.baseHref = "/";
        }
        this.processNavigations();
    }
    /**
     * @return {?}
     */
    get queryParamsSnapshot() {
        return this._queryParams.getValue();
    }
    /**
     * @return {?}
     */
    get queryParams() {
        return this._queryParams;
    }
    /**
     * @return {?}
     */
    get events() {
        return this._events;
    }
    /**
     * @param {?} url
     * @return {?}
     */
    static stripTrailingSlash(url) {
        return url[url.length - 1] === "/" ? url.slice(0, -1) : url;
    }
    /**
     * @param {?} url1
     * @param {?} url2
     * @return {?}
     */
    static urlEquals(url1, url2) {
        url1 = Router.stripTrailingSlash(url1);
        url2 = Router.stripTrailingSlash(url2);
        return decodeURIComponent(url1) === decodeURIComponent(url2);
    }
    /**
     * @return {?}
     */
    initialNavigation() {
        this.setUpLocationChangeListener();
        if (this.navigationId === 0) {
            this.navigateByUrl(this.location.path(true), { replaceUrl: true });
        }
    }
    /**
     * @param {?} url
     * @param {?=} extras
     * @param {?=} params
     * @return {?}
     */
    navigateByUrl(url, extras, params) {
        // absolute url
        /** @type {?} */
        const requestedUrl = UrlParser.parseUrl(url);
        if (url.indexOf("https://") === 0 || url.indexOf("http://") === 0 || url.indexOf("//") === 0) {
            if (window.location
                && window.location.host == requestedUrl.host
                && window.location.protocol == requestedUrl.protocol) {
                // это наш хост, выбросим часть маршрута
                url = requestedUrl.pathname + requestedUrl.search + requestedUrl.hash;
            }
            else {
                throw new Error('Router: cannot navigate to external url');
            }
        }
        else if (url[0] !== "/") {
            // add base href
            url = this.baseHref + url;
        }
        if (params) {
            /** @type {?} */
            let _combinedQueryParams;
            if (requestedUrl.search && requestedUrl.search.length) {
                _combinedQueryParams = Object.assign({}, QueryStringParser.parse(requestedUrl.search), params);
            }
            else {
                _combinedQueryParams = params;
            }
            url = url.split("?", 1)[0];
            url += `?${QueryStringParser.serialize(_combinedQueryParams)}`;
        }
        if ((extras && !extras.force) && this.currentContext && url === this.currentContext.url) {
            // navigation within same url
            return Promise.resolve(true);
        }
        // schedule
        return this.navigateInternal(url, extras);
    }
    /**
     * @return {?}
     */
    reload() {
        if (!this.currentContext) {
            return Promise.resolve(false);
        }
        return this.navigateByUrl(this.currentContext.url, { replaceUrl: true, force: true });
    }
    /**
     * @param {?} p
     * @param {?=} navigationExtras
     * @return {?}
     */
    setQuery(p, navigationExtras) {
        if (this.navigating) {
            return Promise.resolve(null);
        }
        /** @type {?} */
        const queryParams = QueryStringParser.serialize(p);
        /** @type {?} */
        const urlParts = UrlParser.parseUrl(this.location.path(true));
        /** @type {?} */
        const newUrl = queryParams.length ? urlParts.pathname + "?" + queryParams : urlParts.pathname;
        return this.navigateByUrl(newUrl, navigationExtras);
    }
    /**
     * @return {?}
     */
    setUpLocationChangeListener() {
        // Zone.current.wrap is needed because of the issue with RxJS scheduler,
        // which does not work properly with zone.js in IE and Safari
        if (!this.locationSubscription) {
            this.locationSubscription = (/** @type {?} */ (this.location.subscribe(Zone.current.wrap((change) => {
                this.navigateByUrl(change["url"], { replaceUrl: true });
            }))));
        }
    }
    /**
     * @param {?} p
     * @param {?=} navigationExtras
     * @return {?}
     */
    updateQuery(p, navigationExtras) {
        /** @type {?} */
        let existingParams = Object.assign({}, this.queryParamsSnapshot, p);
        return this.setQuery(existingParams, navigationExtras);
    }
    /**
     * @param {?} currentRoute
     * @param {?} injector
     * @return {?}
     */
    checkDeactivate(currentRoute, injector) {
        if (!currentRoute.canDeactivate) {
            return Promise.resolve(true);
        }
        /** @type {?} */
        const outlet = this.outletMap.getOutlet(currentRoute.outlet);
        /** @type {?} */
        const results = [];
        for (let guardType of currentRoute.canDeactivate) {
            /** @type {?} */
            const guard = (/** @type {?} */ (injector.get(guardType)));
            results.push(guard.canDeactivate(outlet.activatedComponent, currentRoute));
        }
        return Promise.all(results).then(all => {
            return !all.some(r => !r);
        });
    }
    /**
     * @param {?} newRoute
     * @return {?}
     */
    checkGuards(newRoute) {
        /** @type {?} */
        let deactivation = this.currentContext ? this.checkDeactivate(this.currentContext.route, this.currentContext.injector) : Promise.resolve(true);
        // check deactivation first
        return deactivation.then(success => {
            if (!success) {
                return { resolvedRoute: newRoute, success: false };
            }
            // check activation
            return checkActivate(newRoute.route, newRoute.injector).then(success => {
                return { resolvedRoute: newRoute, success: success };
            });
        });
    }
    /**
     * @param {?} __0
     * @return {?}
     */
    executeScheduledNavigation({ extras, id, url }) {
        /** @type {?} */
        const force = extras && extras.force;
        if (this.currentContext && (!force && Router.urlEquals(url, this.currentContext.url))) {
            // drop same navigation
            return Promise.resolve(true);
        }
        /** @type {?} */
        const shouldReplaceUrl = extras ? extras.replaceUrl : false;
        /** @type {?} */
        const config = this.injector.get(ROUTE_CONFIG);
        this.navigating = true;
        this._events.next(new NavigationStart());
        /** @type {?} */
        let navigationCanceled;
        /** @type {?} */
        let shouldActivate = true;
        /** @type {?} */
        let outlet;
        return this.matchRoute(url, config)
            .then(route => this.resolveRoute(url, route))
            .then((resolvedRoute) => {
            if (this.navigationId !== id) {
                // navigation changes while resolving
                navigationCanceled = true;
                return undefined;
            }
            if (this.currentContext && this.currentContext.route.path === resolvedRoute.route.path) {
                // navigating with same route
                shouldActivate = force;
                this.currentContext.update(url, resolvedRoute);
                return { resolvedRoute, success: true };
            }
            // set url params to match its parsed state
            if (resolvedRoute.url !== url) {
                url = resolvedRoute.url;
            }
            // check guards if routes changes
            return this.checkGuards(resolvedRoute);
        })
            .then((checkGuardsResult) => {
            if (this.navigationId !== id) {
                // navigation changes while resolving or guard checking
                navigationCanceled = true;
                return this.currentContext;
            }
            if (checkGuardsResult.success) {
                // we should change location state before activate outlet
                if (shouldReplaceUrl) {
                    this.location.replaceState(url);
                }
                else {
                    this.location.go(url);
                }
                outlet = this.outletMap.getOutlet(checkGuardsResult.resolvedRoute.route.outlet);
                /** @type {?} */
                const resolvedRoute = checkGuardsResult.resolvedRoute;
                if (shouldActivate) {
                    // new route navigation
                    return outlet.activate(resolvedRoute, url, resolvedRoute.component, resolvedRoute.factoryResolver, resolvedRoute.injector, force).then(result => result.routeContext);
                }
            }
            else {
                // guard blocked navigation
                this.resetUrlToCurrent();
                navigationCanceled = true;
            }
            return this.currentContext;
        })
            .then((routeContext) => {
            this.currentContext = routeContext;
        })
            .then(() => {
            this.navigating = false;
            if (navigationCanceled) {
                this._events.next(new NavigationCancel());
                return;
            }
            // make sure queryParams changed before pushing to observable
            if (!isParamsEqual(this.queryParamsSnapshot, this.currentContext.queryParamsSnapshot)) {
                this._queryParams.next(this.currentContext.queryParamsSnapshot);
            }
            this._events.next(new NavigationEnd(url, outlet.activatedComponent, this.currentContext));
        })
            .catch((e) => {
            this.navigating = false;
            throw e;
        });
    }
    /**
     * @param {?} url
     * @param {?} config
     * @return {?}
     */
    matchRoute(url, config) {
        /** @type {?} */
        const path = UrlParser.parseUrl(url).pathname;
        if (!this.resolvedRoutes[path]) {
            return this.matchService.findRoute(url, config).then(route => {
                if (!route.noCache) {
                    this.resolvedRoutes[path] = route;
                }
                return route;
            });
        }
        else {
            return Promise.resolve(this.resolvedRoutes[path]);
        }
    }
    /**
     * @param {?} url
     * @param {?} extras
     * @return {?}
     */
    navigateInternal(url, extras) {
        /** @type {?} */
        let resolve = null;
        /** @type {?} */
        let reject = null;
        /** @type {?} */
        const result = new Promise((r, rej) => {
            resolve = r;
            reject = rej;
        });
        this.scheduleNavigation({
            extras: extras,
            id: ++this.navigationId,
            promise: result,
            resolve,
            reject,
            url: url
        });
        return result;
    }
    /**
     * @return {?}
     */
    processNavigations() {
        this.navigations.pipe(concatMap((nav) => {
            if (nav) {
                return this.executeScheduledNavigation(nav).then(nav.resolve, nav.reject);
            }
            else {
                return (/** @type {?} */ (of(null)));
            }
        })).subscribe(() => {
        });
        // initial query string
        /** @type {?} */
        const queryString = UrlParser.parseUrl(this.location.path(true)).search;
        this._queryParams = new BehaviorSubject(QueryStringParser.parse(queryString));
    }
    /**
     * @return {?}
     */
    resetUrlToCurrent() {
        if (this.currentContext) {
            this.location.replaceState(this.currentContext.url);
        }
    }
    /**
     * @param {?} originalUrl
     * @param {?} route
     * @return {?}
     */
    resolveRoute(originalUrl, route) {
        /** @type {?} */
        let urlParts = UrlParser.parseUrl(originalUrl);
        /** @type {?} */
        let queryParams = {};
        if (urlParts.search.length > 0) {
            // set url params to match its parsed state
            queryParams = QueryStringParser.parse(urlParts.search);
            /** @type {?} */
            const qs = QueryStringParser.serialize(queryParams);
            originalUrl = urlParts.pathname + (qs.length ? "?" + qs : "");
        }
        return {
            component: route.component,
            params: route.params,
            factoryResolver: route.factoryResolver,
            injector: route.injector || this.injector,
            route: route.route,
            queryParams,
            url: originalUrl
        };
    }
    /**
     * @param {?} nav
     * @return {?}
     */
    scheduleNavigation(nav) {
        this.navigations.next(nav);
    }
}
Router.decorators = [
    { type: Injectable }
];
/** @nocollapse */
Router.ctorParameters = () => [
    { type: RouterOutletMap },
    { type: RouteMatchService },
    { type: Location },
    { type: Injector },
    { type: String, decorators: [{ type: Optional }, { type: Inject, args: [APP_BASE_HREF,] }] }
];
/**
 * @param {?} newRoute
 * @param {?} injector
 * @return {?}
 */
function checkActivate(newRoute, injector) {
    if (!newRoute.canActivate) {
        return Promise.resolve(true);
    }
    /** @type {?} */
    const results = [];
    for (let guardType of newRoute.canActivate) {
        /** @type {?} */
        const guard = (/** @type {?} */ (injector.get(guardType)));
        results.push(guard.canActivate(newRoute));
    }
    return Promise.all(results).then(all => {
        return !all.some(r => !r);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class RouteContext {
    /**
     * @param {?} url
     * @param {?} resolvedRoute
     */
    constructor(url, resolvedRoute) {
        this._active = true;
        this.update(url, resolvedRoute);
    }
    /**
     * @return {?}
     */
    get queryParams() {
        return this._queryParams;
    }
    /**
     * @return {?}
     */
    get outlet() {
        return this._route.outlet;
    }
    /**
     * @return {?}
     */
    get routeParams() {
        return this._routeParams;
    }
    /**
     * @return {?}
     */
    get url() {
        return this._url;
    }
    /**
     * @return {?}
     */
    get route() {
        return this._route;
    }
    /**
     * @return {?}
     */
    get injector() {
        return this._injector;
    }
    /**
     * @return {?}
     */
    get active() {
        return this._active;
    }
    /**
     * @return {?}
     */
    get queryParamsSnapshot() {
        return this._queryParams.getValue();
    }
    /**
     * @return {?}
     */
    get routeParamsSnapshot() {
        return this._routeParams.getValue();
    }
    /**
     * @return {?}
     */
    activate() {
        this._active = true;
    }
    /**
     * @return {?}
     */
    deactivate() {
        this._active = false;
    }
    /**
     * @param {?} url
     * @param {?} resolvedRoute
     * @return {?}
     */
    update(url, resolvedRoute) {
        //try to keep urls consistent
        this._url = url[0] !== "/" ? "/" + url : url; // TODO: Base Href вместо / ??
        if (!this._queryParams) {
            this._queryParams = new BehaviorSubject(resolvedRoute.queryParams); // TODO query params ??
        }
        else if (!isParamsEqual(this.queryParamsSnapshot, resolvedRoute.queryParams)) {
            this._queryParams.next(resolvedRoute.queryParams); // TODO query params ??
        }
        if (!this._routeParams) {
            this._routeParams = new BehaviorSubject(resolvedRoute.params);
        }
        else if (!isParamsEqual(this.routeParamsSnapshot, resolvedRoute.params, true)) {
            this._routeParams.next(resolvedRoute.params);
        }
        this._route = resolvedRoute.route;
        this._injector = resolvedRoute.injector;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class RouterScrollWrapper {
    /**
     * @return {?}
     */
    getScrollState() {
        return window.pageYOffset || document.documentElement.scrollTop;
    }
    /**
     * @param {?} position
     * @return {?}
     */
    setScrollState(position) {
        window.scrollTo(0, position);
    }
    /**
     * @return {?}
     */
    moveTop() {
        this.setScrollState(0);
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const REUSE_CACHE_CONSTRAINT = new InjectionToken("Router reuse cache constraint");
class RouteReuseCache {
    /**
     * @param {?} router
     * @param {?} constraint
     * @param {?} location
     * @param {?} scrollWrapper
     */
    constructor(router, constraint, location, scrollWrapper) {
        this.router = router;
        this.location = location;
        this.cache = window["_cache"] = [];
        this.newRouteNavigations = [];
        this.constraint = (/** @type {?} */ (constraint));
        if (location.getCurrentPageId) {
            router.events.subscribe((e) => {
                if (e instanceof NavigationEnd && !e.routeContext.outlet) {
                    this.currentCacheItem = this.ensureCache(e.activatedComponent, e.routeContext);
                    if (!this.lastRouteConfig || this.lastRouteConfig !== e.routeContext.route) {
                        this.newRouteNavigation();
                        this.lastRouteConfig = e.routeContext.route;
                    }
                }
                else if (e instanceof NavigationStart) {
                    if (this.currentCacheItem) {
                        this.currentCacheItem.scrollState = scrollWrapper.getScrollState();
                    }
                }
            });
        }
        else {
            console.warn(`Router will not reuse routes because your app use custom LocationStrategy which not provide information about page id`);
        }
    }
    /**
     * @param {?} ref
     * @return {?}
     */
    getCacheFor(ref) {
        return this.cache.find(c => c.ref === ref);
    }
    /**
     * @param {?} cmpType
     * @return {?}
     */
    getForCurrentPage(cmpType) {
        /** @type {?} */
        const cacheItems = this.cache.filter(c => c.ref.componentType === cmpType);
        if (!cacheItems.length) {
            return undefined; // no cache
        }
        /** @type {?} */
        const pageId = this.getCurrentPageId();
        for (let cacheItem of cacheItems) {
            if (cacheItem.pageIds) {
                //cache back
                if (cacheItem.hasPageId(pageId)) {
                    return cacheItem;
                }
                continue;
            }
            return cacheItem; // sticky
        }
        return undefined;
    }
    /**
     * @param {?} component
     * @param {?} routeCtx
     * @return {?}
     */
    ensureCache(component, routeCtx) {
        if (!component.instance.reuseRouteStrategy) {
            return undefined; // not a reusable component
        }
        // do not cache other outlets than default
        /** @type {?} */
        const cacheBack = component.instance.reuseRouteStrategy === ReuseRouteStrategy.CACHEBACK && !routeCtx.outlet;
        /** @type {?} */
        const pageId = this.getCurrentPageId();
        if (!pageId) {
            return undefined; // history api may not supoprted
        }
        /** @type {?} */
        let result;
        if (cacheBack) {
            /** @type {?} */
            let existCacheItems = this.cache.filter(c => c.ref.componentType === component.componentType);
            for (let cacheItem of existCacheItems) {
                if (cacheItem.ref.instance === component.instance) {
                    cacheItem.addPageId(pageId);
                    return cacheItem;
                }
            }
            result = new ReuseCacheItem(component, routeCtx, pageId);
            this.cache.push(result);
        }
        else {
            /** @type {?} */
            const existsSticky = this.cache.some(c => c.ref.componentType === component.componentType);
            if (!existsSticky) {
                result = new ReuseCacheItem(component, routeCtx);
                this.cache.push(result);
            }
        }
        return result;
    }
    /**
     * @return {?}
     */
    getCurrentPageId() {
        return this.location.getCurrentPageId();
    }
    /**
     * @return {?}
     */
    newRouteNavigation() {
        // removes cache which out of constraint
        this.newRouteNavigations.push(this.getCurrentPageId());
        if (this.newRouteNavigations.length > this.constraint.maxHistoryCacheDepth + 1 /* include current navigation */) {
            this.newRouteNavigations.shift();
        }
        /** @type {?} */
        let cacheItemsToDrop = [];
        /** @type {?} */
        let stickyRoutesUsed = 0;
        for (const cacheItem of this.cache) {
            // do not clear cacheback for all other outlets
            if (cacheItem.reuseStrategy === ReuseRouteStrategy.CACHEBACK && !cacheItem.routeContext.outlet) {
                if (this.newRouteNavigations.every(n => !cacheItem.hasPageId(n))) {
                    cacheItemsToDrop.push(cacheItem);
                }
            }
            else {
                stickyRoutesUsed++;
                if (stickyRoutesUsed > this.constraint.maxStickyRoutes) {
                    cacheItemsToDrop.push(cacheItem);
                }
            }
        }
        for (let cacheToDrop of cacheItemsToDrop) {
            cacheToDrop.destroy();
            this.cache.splice(this.cache.indexOf(cacheToDrop), 1);
        }
    }
}
RouteReuseCache.decorators = [
    { type: Injectable }
];
/** @nocollapse */
RouteReuseCache.ctorParameters = () => [
    { type: Router },
    { type: undefined, decorators: [{ type: Inject, args: [REUSE_CACHE_CONSTRAINT,] }] },
    { type: undefined, decorators: [{ type: Inject, args: [LocationStrategy,] }] },
    { type: RouterScrollWrapper }
];
class ReuseCacheItem {
    /**
     * @param {?} ref
     * @param {?} routeCtx
     * @param {?=} pageId
     */
    constructor(ref, routeCtx, pageId) {
        if (pageId) {
            this._pageIds = [pageId];
        }
        this._ref = ref;
        this._routeCtx = routeCtx;
    }
    /**
     * Cached within specified pageIds
     * @return {?}
     */
    get pageIds() {
        return this._pageIds;
    }
    /**
     * Cached component reference
     * @return {?}
     */
    get ref() {
        return this._ref;
    }
    /**
     * @return {?}
     */
    get routeContext() {
        return this._routeCtx;
    }
    /**
     * @return {?}
     */
    get reuseStrategy() {
        return this._ref.instance.reuseRouteStrategy;
    }
    /**
     * @param {?} pageId
     * @return {?}
     */
    addPageId(pageId) {
        if (!this._ref) {
            throw new Error("Route cache item was destroyed");
        }
        if (!this._pageIds) {
            this._pageIds = [];
        }
        if (this.hasPageId(pageId)) {
            return;
        }
        this._pageIds.push(pageId);
    }
    /**
     * @return {?}
     */
    attached() {
        if (!this._ref) {
            throw new Error("Route cache item was destroyed");
        }
        if (this.ref.instance.onRouteReused) {
            this.ref.instance.onRouteReused();
        }
        this.routeContext.activate();
    }
    /**
     * @return {?}
     */
    destroy() {
        if (this._ref) {
            this._ref.destroy();
            this._ref = null;
        }
    }
    /**
     * @return {?}
     */
    detached() {
        if (!this._ref) {
            throw new Error("Route cache item was destroyed");
        }
        if (this.ref.instance.onRouteCached) {
            this.ref.instance.onRouteCached();
        }
        this.routeContext.deactivate();
    }
    /**
     * @param {?} pageId
     * @return {?}
     */
    hasPageId(pageId) {
        if (!this._pageIds) {
            return false;
        }
        return this._pageIds.indexOf(pageId) >= 0;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
class RouterExOutletComponent {
    /**
     * @param {?} parentOutletMap
     * @param {?} location
     * @param {?} reuseCache
     * @param {?} scrollWrapper
     * @param {?} name
     */
    constructor(parentOutletMap, location, reuseCache, scrollWrapper, name) {
        this.parentOutletMap = parentOutletMap;
        this.location = location;
        this.reuseCache = reuseCache;
        this.scrollWrapper = scrollWrapper;
        this.name = name;
        this.prerenderFallback = 500;
        this.activateEvents = new EventEmitter();
        this.deactivateEvents = new EventEmitter();
        this.prerenderReadyEvents = new EventEmitter();
        parentOutletMap.register((/** @type {?} */ (this)), name);
    }
    /**
     * @return {?}
     */
    get isActivated() {
        return !!this.activated;
    }
    /**
     * @return {?}
     */
    get activatedComponent() {
        return this.activated;
    }
    /**
     * @param {?} route
     * @param {?} url
     * @param {?} componentType
     * @param {?} resolver
     * @param {?} injector
     * @param {?=} force
     * @return {?}
     */
    activate(route, url, componentType, resolver, injector, force) {
        if (!force && this.routeCtx && routeEquals(this.routeCtx.route, route.route)) {
            this.routeCtx.update(url, route);
            return Promise.resolve({ routeContext: this.routeCtx });
        }
        /** @type {?} */
        const cache = this.reuseCache.getForCurrentPage(componentType);
        if (cache) {
            // reuse existing component ref
            this.activateComponent(cache.ref, cache.routeContext);
            cache.attached();
            cache.ref.changeDetectorRef.detectChanges();
            if (cache.reuseStrategy === ReuseRouteStrategy.CACHEBACK) {
                this.scrollWrapper.setScrollState(cache.scrollState);
            }
            cache.routeContext.update(url, route);
            return Promise.resolve({
                routeContext: cache.routeContext
            });
        }
        /** @type {?} */
        const routeContext = new RouteContext(url, route);
        // create new instance of the component
        /** @type {?} */
        const factory = resolver.resolveComponentFactory(componentType);
        /** @type {?} */
        const inj = Injector.create([
            { provide: RouteContext, useValue: routeContext }
        ], injector);
        //const inj = bindings.get(RouteContext)//= ReflectiveInjector.fromResolvedProviders(bindings, injector);
        /** @type {?} */
        let componentToActivate = this.prerenderContainer.createComponent(factory, this.prerenderContainer.length, inj, []);
        /** @type {?} */
        const onDone = () => {
            if (this.autoScroll) {
                this.scrollWrapper.moveTop();
            }
        };
        /** @type {?} */
        const noPrerender = () => {
            // no prerender
            this.prerenderContainer.detach();
            this.activateComponent(componentToActivate, routeContext);
            onDone();
            return Promise.resolve({
                routeContext
            });
        };
        // need to prerender component ?
        /** @type {?} */
        const prerenderComponent = (/** @type {?} */ (componentToActivate.instance));
        if (prerenderComponent.routePrerender) {
            /** @type {?} */
            let rendered = false;
            /** @type {?} */
            const routeReady = ((/** @type {?} */ (componentToActivate.instance)))
                .routePrerender()
                .then(() => {
                this.prerenderReadyEvents.emit(componentToActivate);
            }, (err) => {
                this.prerenderReadyEvents.emit(componentToActivate);
                throw err;
            });
            if (this.activatedComponent) {
                return new Promise(resolve => {
                    /** @type {?} */
                    const delay$ = of(true)
                        .pipe(delay(prerenderComponent.fallbackTimeout || this.prerenderFallback));
                    from(routeReady)
                        .pipe(merge(delay$))
                        .subscribe(() => {
                        if (rendered) {
                            return;
                        }
                        rendered = true;
                        this.prerenderContainer.detach();
                        this.activateComponent(componentToActivate, routeContext);
                        onDone();
                        resolve({ routeContext });
                    });
                });
            }
            else {
                return noPrerender();
            }
        }
        else {
            return noPrerender();
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.parentOutletMap.unregister(this.name);
    }
    /**
     * Replace current activated component with specified one
     * @param {?} ref
     * @param {?} routeCtx
     * @return {?}
     */
    activateComponent(ref, routeCtx) {
        this.deactivateCurrent();
        this.attach(ref, routeCtx);
        this.activateEvents.emit(this.activated.instance);
        if (ref.instance.onRouteOutletActivated) {
            ref.instance.onRouteOutletActivated();
        }
    }
    /**
     * @param {?} ref
     * @param {?} routeContext
     * @return {?}
     */
    attach(ref, routeContext) {
        this.activated = ref;
        this.routeCtx = routeContext;
        this.location.insert(ref.hostView);
    }
    /**
     * @return {?}
     */
    deactivateCurrent() {
        if (this.activated) {
            /** @type {?} */
            const c = this.activated.instance;
            /** @type {?} */
            const cache = this.reuseCache.getCacheFor(this.activated);
            if (this.activated.instance.onRouteOutletDeactivated) {
                this.activated.instance.onRouteOutletDeactivated();
            }
            if (cache) {
                this.location.detach();
                cache.detached();
            }
            else {
                this.activated.destroy();
            }
            this.activated = null;
            this.deactivateEvents.emit(c);
        }
    }
}
RouterExOutletComponent.decorators = [
    { type: Component, args: [{
                selector: "router-ex-outlet",
                template: `
        <div #placeholder></div>`,
                styles: [`:host {
        display: none;
    }`]
            }] }
];
/** @nocollapse */
RouterExOutletComponent.ctorParameters = () => [
    { type: RouterOutletMap },
    { type: ViewContainerRef },
    { type: RouteReuseCache },
    { type: RouterScrollWrapper },
    { type: String, decorators: [{ type: Attribute, args: ["name",] }] }
];
RouterExOutletComponent.propDecorators = {
    autoScroll: [{ type: Input }],
    prerenderFallback: [{ type: Input }],
    activateEvents: [{ type: Output, args: ["onActivate",] }],
    deactivateEvents: [{ type: Output, args: ["onDeactivate",] }],
    prerenderReadyEvents: [{ type: Output, args: ["prerenderReady",] }],
    prerenderContainer: [{ type: ViewChild, args: ["placeholder", { read: ViewContainerRef },] }]
};
/**
 * @param {?} route
 * @param {?} route2
 * @return {?}
 */
function routeEquals(route, route2) {
    return route.path === route2.path;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * Intercept all document click on anchor href and perform router navigation
 */
class HandleHrefNavigationDirective {
    /**
     * @param {?} router
     */
    constructor(router) {
        this.router = router;
        this.handler = (event) => {
            /*
             *  Взято из первого ангуляра
             */
            if (event.defaultPrevented) {
                return;
            }
            /* открытие в новом окне и т.п. не обрабатываем */
            if (event.ctrlKey || event.metaKey || event.shiftKey || event.which === 2 || event.button === 2)
                return;
            /** @type {?} */
            let elm = (/** @type {?} */ (event.target));
            // traverse the DOM up to find first A tag
            while (!elm.nodeName || elm.nodeName.toUpperCase() !== "A") {
                // ignore rewriting if no A tag (reached root element, or no parent - removed from document)
                if (elm === document.body || !(elm = elm.parentElement))
                    return;
            }
            /** @type {?} */
            let navigateUrl = elm.getAttribute("href");
            /* не обрабатываем пустые ссылки,
             ссылки, открывающиеся в новом окне,
             ссылки на внешние ресурсы, а также ссылки с url-scheme приложений и на внешние url*/
            if (navigateUrl &&
                (elm.getAttribute("target") === "_blank" || /^\s*(javascript|mailto|tel|sip):/i.test(navigateUrl)) ||
                HandleHrefNavigationDirective.isExternalNavigation(navigateUrl)) {
                return;
            }
            event.preventDefault();
            if (navigateUrl) {
                this.router.navigateByUrl(navigateUrl);
            }
        };
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        document.body.addEventListener("click", this.handler);
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        document.body.removeEventListener("click", this.handler);
    }
    /**
     * @param {?} url
     * @return {?}
     */
    static isExternalNavigation(url) {
        /** @type {?} */
        const parsed = UrlParser.parseUrl(url);
        return isAbsUrl(url) && window.location.host !== parsed.host;
    }
}
HandleHrefNavigationDirective.decorators = [
    { type: Directive, args: [{
                selector: "[handleHrefNavigation]"
            },] }
];
/** @nocollapse */
HandleHrefNavigationDirective.ctorParameters = () => [
    { type: Router }
];
/**
 * @param {?} url
 * @return {?}
 */
function isAbsUrl(url) {
    if (!url) {
        return false;
    }
    return (url.indexOf("https://") === 0 || url.indexOf("http://") === 0 || url.indexOf("//") === 0);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/**
 * Состояние, которое сохраняется в HISTORY STATE
 *
 * \@property {boolean} pageId Уникальный идентификатор страницы в истории браузера
 */
class HistoryPageState {
    /**
     * @param {?} pageId
     */
    constructor(pageId) {
        this.pageId = pageId;
    }
    /**
     * Создает новое состояние с уникальным pageID
     * @return {?}
     */
    static createNew() {
        return new HistoryPageState(generateUniqueId());
    }
}
class HistoryApiLocationStrategy extends LocationStrategy {
    /**
     * @param {?} _platformLocation
     * @param {?=} href
     */
    constructor(_platformLocation, href) {
        super();
        this._platformLocation = _platformLocation;
        if (href == null) {
            href = this._platformLocation.getBaseHrefFromDOM();
        }
        if (href == null) {
            console.warn(`No base href set. Please provide a value for the APP_BASE_HREF token or add a base element to the document.`);
            href = "/"; // TODO BUG после обновления до 4.0.0-rc.4 почему то перестал инджектится APP_BASE_HREF
        }
        this._baseHref = href;
    }
    /**
     * @return {?}
     */
    getCurrentPageId() {
        return history.state ? history.state.pageId : undefined;
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    onPopState(fn) {
        this._platformLocation.onPopState(fn);
        this._platformLocation.onHashChange(fn);
    }
    /**
     * @return {?}
     */
    getBaseHref() {
        return this._baseHref;
    }
    /**
     * @param {?} internal
     * @return {?}
     */
    prepareExternalUrl(internal) {
        return Location.joinWithSlash(this._baseHref, internal);
    }
    /**
     * @param {?=} includeHash
     * @return {?}
     */
    path(includeHash = false) {
        /** @type {?} */
        const pathname = this._platformLocation.pathname +
            Location.normalizeQueryParams(this._platformLocation.search);
        /** @type {?} */
        const hash = this._platformLocation.hash;
        return hash && includeHash ? `${pathname}${hash}` : pathname;
    }
    /**
     * @param {?} state
     * @param {?} title
     * @param {?} url
     * @param {?} queryParams
     * @return {?}
     */
    pushState(state, title, url, queryParams) {
        /** @type {?} */
        const externalUrl = this.prepareExternalUrl(url + Location.normalizeQueryParams(queryParams));
        this._platformLocation.pushState(HistoryPageState.createNew(), title, externalUrl);
        //resetPopedState();
    }
    /**
     * @param {?} state
     * @param {?} title
     * @param {?} url
     * @param {?} queryParams
     * @return {?}
     */
    replaceState(state, title, url, queryParams) {
        /** @type {?} */
        const externalUrl = this.prepareExternalUrl(url + Location.normalizeQueryParams(queryParams));
        this._platformLocation.replaceState(ensureHistoryState(), title, externalUrl);
        //resetPopedState();
    }
    /**
     * @return {?}
     */
    forward() {
        this._platformLocation.forward();
    }
    /**
     * @return {?}
     */
    back() {
        this._platformLocation.back();
    }
}
HistoryApiLocationStrategy.decorators = [
    { type: Injectable }
];
/** @nocollapse */
HistoryApiLocationStrategy.ctorParameters = () => [
    { type: PlatformLocation },
    { type: String, decorators: [{ type: Optional }, { type: Inject, args: [APP_BASE_HREF,] }] }
];
/**
 * @param {?=} state
 * @return {?}
 */
function ensureHistoryState(state) {
    /** @type {?} */
    let currentState = state || window.history.state;
    if (!currentState) {
        currentState = HistoryPageState.createNew();
    }
    return currentState;
}
/**
 * @return {?}
 */
function generateUniqueId() {
    /** @type {?} */
    let idstr = String.fromCharCode(Math.floor((Math.random() * 25) + 65));
    do {
        // between numbers and characters (48 is 0 and 90 is Z (42-48 = 90)
        /** @type {?} */
        let ascicode = Math.floor((Math.random() * 42) + 48);
        if (ascicode < 58 || ascicode > 64) {
            // exclude all chars between : (58) and @ (64)
            idstr += String.fromCharCode(ascicode);
        }
    } while (idstr.length < 5);
    return idstr;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */
/** @type {?} */
const DEFAULT_CACHE_CONSTRAINT = { maxStickyRoutes: 2, maxHistoryCacheDepth: 1 };
class RouterExModule {
    /**
     * @param {?} routeConfig
     * @return {?}
     */
    static forRoot(routeConfig) {
        return {
            ngModule: RouterExModule,
            providers: [
                Location,
                RouterOutletMap,
                RouteMatchService,
                Router,
                RouteReuseCache,
                RouterScrollWrapper,
                { provide: RouterModuleFactory, useClass: DefaultRouterModuleFactory },
                { provide: ANALYZE_FOR_ENTRY_COMPONENTS, multi: true, useValue: routeConfig },
                { provide: NgModuleFactoryLoader, useClass: SystemJsNgModuleLoader },
                { provide: LocationStrategy, useClass: HistoryApiLocationStrategy },
                { provide: ROUTE_CONFIG, useValue: routeConfig },
                {
                    provide: REUSE_CACHE_CONSTRAINT,
                    useValue: DEFAULT_CACHE_CONSTRAINT
                },
                {
                    provide: APP_BOOTSTRAP_LISTENER, multi: true, useFactory: initialRouterNavigation,
                    deps: [Router, ApplicationRef]
                }
            ]
        };
    }
    /**
     * @param {?} routeConfig
     * @return {?}
     */
    static forChild(routeConfig) {
        return {
            ngModule: RouterExModule,
            providers: [
                { provide: ROUTE_CONFIG, useValue: routeConfig },
                { provide: ANALYZE_FOR_ENTRY_COMPONENTS, multi: true, useValue: routeConfig },
            ]
        };
    }
}
RouterExModule.decorators = [
    { type: NgModule, args: [{
                exports: [
                    RouterExOutletComponent,
                    HandleHrefNavigationDirective
                ],
                declarations: [
                    RouterExOutletComponent,
                    HandleHrefNavigationDirective
                ],
                providers: [
                    QueryStringParser
                ],
                imports: [
                    CommonModule
                ]
            },] }
];
/**
 * @param {?} router
 * @param {?} ref
 * @return {?}
 */
function initialRouterNavigation(router, ref) {
    /**
     * @param {?} bootstrappedComponentRef
     * @return {?}
     */
    function initNav(bootstrappedComponentRef) {
        if (bootstrappedComponentRef !== ref.components[0]) {
            return;
        }
        router.initialNavigation();
    }
    return initNav;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,uselessCode} checked by tsc
 */

export { Router, RouterOutletMap, ReuseRouteStrategy, RouterExModule, NavigationCancel, NavigationEnd, NavigationStart, RouteContext, RouterModuleFactory, UrlParser, ROUTE_CONFIG as ɵa, HistoryApiLocationStrategy as ɵi, QueryStringParser as ɵl, RouteMatchService as ɵd, REUSE_CACHE_CONSTRAINT as ɵf, RouteReuseCache as ɵg, DefaultRouterModuleFactory as ɵc, RouterScrollWrapper as ɵj, HandleHrefNavigationDirective as ɵk, RouterExOutletComponent as ɵe, initialRouterNavigation as ɵb };

//# sourceMappingURL=angular-router-ex.js.map